import logging
from pathlib import Path

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QHBoxLayout,
    QLineEdit,
    QFormLayout,
    QLabel,
    QRadioButton,
    QCheckBox,
    QComboBox,
    QFrame,
)
from skellycam import SkellyCamControllerWidget, SkellyCamWidget

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import CHARUCO_BOARDS
from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState, load_gui_state, save_gui_state
from freemocap.gui.qt.widgets.calibration_guide_link_qlabel import CalibrationGuideLinkQLabel
from freemocap.system.paths_and_filenames.file_and_folder_names import SPARKLES_EMOJI_STRING, SKULL_EMOJI_STRING
from freemocap.system.paths_and_filenames.path_getters import (
    create_new_recording_folder_path,
    create_new_default_recording_name,
    get_gui_state_json_path,
)

CALIBRATION_RECORDING_BUTTON_TEXT = "\U0001f534 \U0001f4d0 Start Calibration Recording"
MOCAP_RECORDING_BUTTON_TEXT = f"{SKULL_EMOJI_STRING} {SPARKLES_EMOJI_STRING} Start Motion Capture Recording"
CALIBRATION_DOCS_LINK = "https://freemocap.github.io/documentation/multi-camera-calibration.html"
logger = logging.getLogger(__name__)


class CameraControllerGroupBox(QGroupBox):
    controller_group_box_calibration_updated = Signal()

    def __init__(self, skellycam_widget: SkellyCamWidget, gui_state: GuiState, parent=None):
        super().__init__(parent=parent)
        self.setStyleSheet("font-size: 12px;")
        self._skellycam_widget = skellycam_widget
        self._skellycam_controller = SkellyCamControllerWidget(
            camera_viewer_widget=skellycam_widget,
            parent=self,
        )
        self._skellycam_controller.check_recording_type = (self.check_recording_type,)

        self.gui_state = gui_state

        self._layout = QHBoxLayout()
        self._layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self.setLayout(self._layout)

        create_calibration_recording_option_layout = self._create_calibration_recording_option_layout()

        vbox = QVBoxLayout()
        vbox.addLayout(self._make_options_layout())
        vbox.addLayout(self._create_mocap_recording_option_layout())
        vbox.addLayout(create_calibration_recording_option_layout)

        self._layout.addLayout(vbox)

        self._calibration_videos_radio_button.toggled.connect(self._set_record_button_text)
        self._annotate_charuco_checkbox.toggled.connect(self._on_annotate_charuco_checkbox_changed)
        self._mocap_videos_radio_button.toggled.connect(self._set_record_button_text)
        self._skellycam_widget.cameras_connected_signal.connect(lambda: self._start_recording_button.setEnabled(True))
        self._stop_recording_button.clicked.connect(self._set_record_button_text)

        self._auto_process_videos_checkbox.toggled.connect(self._on_auto_process_videos_checkbox_changed)
        self._generate_jupyter_notebook_checkbox.toggled.connect(self._on_generate_jupyter_notebook_checkbox_changed)
        self._auto_open_in_blender_checkbox.toggled.connect(self._on_auto_open_in_blender_checkbox_changed)
        self._charuco_square_size_line_edit.textEdited.connect(self._on_charuco_square_size_line_edit_changed)
        self._use_charuco_as_groundplane_checkbox.toggled.connect(self._on_use_charuco_groundplane_checkbox_changed)
        self._board_dropdown.currentTextChanged.connect(self._on_charuco_board_dropdown_changed)

    @property
    def mocap_videos_radio_button_checked(self) -> bool:
        return self._mocap_videos_radio_button.isChecked()

    @property
    def calibration_videos_radio_button_checked(self) -> bool:
        return self._calibration_videos_radio_button.isChecked()

    @property
    def auto_process_videos_checked(self) -> bool:
        return self._auto_process_videos_checkbox.isChecked()

    @property
    def auto_open_in_blender_checked(self) -> bool:
        return self._auto_open_in_blender_checkbox.isChecked()

    @property
    def generate_jupyter_notebook_checked(self) -> bool:
        return self._generate_jupyter_notebook_checkbox.isChecked()

    @property
    def charuco_square_size(self) -> float:
        return float(self._charuco_square_size_line_edit.text())

    @property
    def use_charuco_as_groundplane(self) -> bool:
        return self._use_charuco_as_groundplane_checkbox.isChecked()

    @property
    def charuco_board_name(self) -> str:
        return self._board_dropdown.currentText()

    def check_recording_type(self) -> str:
        if self._mocap_videos_radio_button.isChecked():
            return "mocap"
        elif self._calibration_videos_radio_button.isChecked():
            return "calibration"
        else:
            raise ValueError("No recording type selected")

    def _create_mocap_recording_option_layout(self) -> QHBoxLayout:
        hbox = QHBoxLayout()
        hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self._mocap_videos_radio_button = QRadioButton("Record Motion Capture Videos")

        hbox.addWidget(self._mocap_videos_radio_button)

        self._mocap_videos_radio_button.setChecked(True)
        hbox.addWidget(QLabel(" - "))
        self._auto_process_videos_checkbox = QCheckBox("Auto Process Videos")
        self._auto_process_videos_checkbox.setChecked(self.gui_state.auto_process_videos_on_save)
        hbox.addWidget(self._auto_process_videos_checkbox)

        self._generate_jupyter_notebook_checkbox = QCheckBox("Generate Jupyter Notebook")
        self._generate_jupyter_notebook_checkbox.setChecked(self.gui_state.generate_jupyter_notebook)
        hbox.addWidget(self._generate_jupyter_notebook_checkbox)

        self._auto_open_in_blender_checkbox = QCheckBox("Auto Open in Blender")
        self._auto_open_in_blender_checkbox.setChecked(self.gui_state.auto_open_in_blender)
        hbox.addWidget(self._auto_open_in_blender_checkbox)
        return hbox

    def _create_calibration_recording_option_layout(self) -> QVBoxLayout:
        vbox = QVBoxLayout()

        # -------- First Row (Calibration Radio, Board Dropdown, Square Size) --------
        hbox_top = QHBoxLayout()
        hbox_top.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self._calibration_videos_radio_button = QRadioButton("Record Calibration Videos")
        hbox_top.addWidget(self._calibration_videos_radio_button)
        hbox_top.addWidget(QLabel(" - "))

        hbox_top.addWidget(QLabel("Charuco square size (mm)", parent=self))
        self._charuco_square_size_line_edit = QLineEdit(parent=self)
        self._charuco_square_size_line_edit.setFixedWidth(60)
        self._charuco_square_size_line_edit.setText(str(self.gui_state.charuco_square_size))
        self._charuco_square_size_line_edit.setToolTip(
            "The length of one of the edges of the black squares in the calibration board in mm"
        )
        hbox_top.addWidget(self._charuco_square_size_line_edit)

        hbox_top.addWidget(QLabel("Charuco Board:"))
        self._board_dropdown = self._create_board_dropdown()
        self._board_dropdown.setCurrentText(self.gui_state.charuco_board_name)
        hbox_top.addWidget(self._board_dropdown)

        hbox_top.addStretch()

        hbox_bottom = QHBoxLayout()
        hbox_bottom.setAlignment(Qt.AlignmentFlag.AlignLeft)
        # Add indentation spacing (adjust the width as needed)
        indent_spacer = QLabel("")
        indent_spacer.setFixedWidth(80)  # Adjust this value for more/less indentation
        hbox_bottom.addWidget(indent_spacer)
        self._annotate_charuco_checkbox = QCheckBox("Charuco Overlay (requires camera restart)")
        self._annotate_charuco_checkbox.setChecked(self.gui_state.annotate_charuco_images)
        self._skellycam_widget.annotate_images = self._annotate_charuco_checkbox.isChecked()
        hbox_bottom.addWidget(self._annotate_charuco_checkbox)

        self._use_charuco_as_groundplane_checkbox = QCheckBox("Use initial board position as origin")
        self._use_charuco_as_groundplane_checkbox.setChecked(self.gui_state.use_charuco_as_groundplane)
        hbox_bottom.addWidget(self._use_charuco_as_groundplane_checkbox)

        hbox_bottom.addWidget(CalibrationGuideLinkQLabel(parent=self))
        # hbox_bottom.addStretch()

        vbox.addLayout(hbox_top)
        vbox.addLayout(hbox_bottom)

        return vbox

    def _open_calibration_docs(self, url: str) -> None:
        """Open the calibration documentation in the default web browser."""
        QDesktopServices.openUrl(url)

    def _make_options_layout(self) -> QHBoxLayout:
        options_vbox = QHBoxLayout()

        options_vbox.addLayout(self._make_record_button_layout())

        options_vbox.addLayout(self._create_videos_will_save_to_layout())

        return options_vbox

    def _create_horizontal_separator(self) -> QFrame:
        """Create a horizontal line separator widget."""
        separator = QFrame()
        separator.setFrameShape(QFrame.Shape.HLine)
        separator.setFrameShadow(QFrame.Shadow.Sunken)
        separator.setLineWidth(1)
        # Optional: Add some styling to make it more visible or subtle
        separator.setStyleSheet(
            """
            QFrame {
                color: #d0d0d0;
                margin-top: 5px;
                margin-bottom: 5px;
            }
        """
        )
        return separator

    def _make_record_button_layout(self) -> QVBoxLayout:
        button_vbox = QVBoxLayout()
        self._start_recording_button = self._skellycam_controller.start_recording_button
        self._stop_recording_button = self._skellycam_controller.stop_recording_button
        self._start_recording_button.setStyleSheet("font-size: 14px;")
        self._stop_recording_button.setStyleSheet("font-size: 14px;")
        self._start_recording_button.setText(MOCAP_RECORDING_BUTTON_TEXT)
        self._start_recording_button.setObjectName("start_recording_button")
        self._stop_recording_button.setObjectName("stop_recording_button")
        self._start_recording_button.show()
        self._start_recording_button.setEnabled(False)
        self._stop_recording_button.show()
        button_vbox.addWidget(self._start_recording_button)
        button_vbox.addWidget(self._stop_recording_button)
        return button_vbox

    def _create_videos_will_save_to_layout(self) -> QVBoxLayout:
        vbox = QVBoxLayout()

        session_path_hbox = QHBoxLayout()
        session_path_hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)
        vbox.addLayout(session_path_hbox)
        recording_path_full = self.get_new_recording_path()
        recording_name = Path(recording_path_full).stem

        recording_name_hbox = QHBoxLayout()
        vbox.addLayout(recording_name_hbox)
        recording_name_key_label = QLabel("Recording Name: ")
        recording_name_hbox.addWidget(recording_name_key_label)
        self._recording_name_label = QLabel(f"{recording_name}")
        self._recording_name_label.setStyleSheet("font-family: monospace;")
        recording_name_hbox.addWidget(self._recording_name_label)

        recording_string_tag_form_layout = QFormLayout(parent=self)
        self._recording_string_tag_line_edit = QLineEdit(parent=self)
        self._recording_string_tag_line_edit.setPlaceholderText("(Optional)")
        self._recording_string_tag_line_edit.setMaxLength(200)
        self._recording_string_tag_line_edit.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        recording_string_tag_form_layout.addRow("Tag:", self._recording_string_tag_line_edit)
        recording_name_hbox.addWidget(QLabel(" - "))
        recording_name_hbox.addLayout(recording_string_tag_form_layout)

        full_path_hbox = QHBoxLayout()
        full_path_hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)
        vbox.addLayout(full_path_hbox)
        full_path_key_label = QLabel("Full Path: ")
        full_path_hbox.addWidget(full_path_key_label)
        self._full_path_label = QLabel(recording_path_full)
        self._full_path_label.setWordWrap(True)
        self._full_path_label.setStyleSheet("font-family: monospace;")
        full_path_hbox.addWidget(self._full_path_label)

        return vbox

    def _create_board_dropdown(self) -> QComboBox:
        board_dropdown = QComboBox()
        board_dropdown.setToolTip("Select the Charuco board to use for calibration")
        board_dropdown.setFixedWidth(150)
        board_dropdown.setStyleSheet("QComboBox { font-size: 12px; }")
        board_dropdown.setEnabled(True)
        board_dropdown.setEditable(False)
        board_dropdown.addItems(list(CHARUCO_BOARDS.keys()))
        return board_dropdown

    def get_new_recording_path(self) -> str:
        return create_new_recording_folder_path(recording_name=self._get_recording_name())

    def update_recording_name_string(self) -> None:
        self._recording_name_label.setText(Path(self.get_new_recording_path()).stem)
        self._full_path_label.setText(f"{self.get_new_recording_path()}")

    def _get_recording_name_string_tag(self) -> str:
        try:
            tag = self._recording_string_tag_line_edit.text()
            tag = tag.replace("   ", " ")
            tag = tag.replace("  ", " ")
            tag = tag.replace(" ", "_")
            return tag
        except Exception:
            return ""

    def _get_recording_name(self) -> str:
        name = create_new_default_recording_name()
        if self._calibration_videos_radio_button.isChecked():
            name = f"{name}_calibration"

        tag = self._get_recording_name_string_tag()
        if tag == "":
            return name
        else:
            return f"{name}__{self._get_recording_name_string_tag()}"

    def _set_record_button_text(self) -> None:
        if self._calibration_videos_radio_button.isChecked():
            self.setProperty("calibration_videos_radio_button_checked", True)
            self._start_recording_button.setText(CALIBRATION_RECORDING_BUTTON_TEXT)
        else:
            self.setProperty("calibration_videos_radio_button_checked", False)
            self._start_recording_button.setText(MOCAP_RECORDING_BUTTON_TEXT)
        self.style().polish(self)

    def _on_annotate_charuco_checkbox_changed(self) -> None:
        self._skellycam_widget.annotate_images = self._annotate_charuco_checkbox.isChecked()
        self.gui_state.annotate_charuco_images = self._annotate_charuco_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

    def _on_use_charuco_groundplane_checkbox_changed(self) -> None:
        self.gui_state.use_charuco_as_groundplane = self._use_charuco_as_groundplane_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.controller_group_box_calibration_updated.emit()

    def _on_auto_process_videos_checkbox_changed(self) -> None:
        self.gui_state.auto_process_videos_on_save = self._auto_process_videos_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

    def _on_generate_jupyter_notebook_checkbox_changed(self) -> None:
        self.gui_state.generate_jupyter_notebook = self._generate_jupyter_notebook_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

    def _on_auto_open_in_blender_checkbox_changed(self) -> None:
        self.gui_state.auto_open_in_blender = self._auto_open_in_blender_checkbox.isChecked()
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())

    def _on_charuco_square_size_line_edit_changed(self) -> None:
        try:
            self.gui_state.charuco_square_size = float(self._charuco_square_size_line_edit.text())
        except ValueError:
            return
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.controller_group_box_calibration_updated.emit()

    def _on_charuco_board_dropdown_changed(self) -> None:
        selected_board_name = self._board_dropdown.currentText()
        self.gui_state.charuco_board_name = selected_board_name
        save_gui_state(gui_state=self.gui_state, file_pathstring=get_gui_state_json_path())
        self.controller_group_box_calibration_updated.emit()

    @Slot()
    def charuco_option_updated(self) -> None:
        self.gui_state = load_gui_state(file_pathstring=get_gui_state_json_path())
        self._board_dropdown.setCurrentText(self.gui_state.charuco_board_name)
        self._charuco_square_size_line_edit.setText(str(self.gui_state.charuco_square_size))
        self._use_charuco_as_groundplane_checkbox.setChecked(self.gui_state.use_charuco_as_groundplane)



from ctypes import Union
from typing import get_origin, get_args

from pydantic import BaseModel


def create_nested_dict(pydantic_model):
    if not issubclass(pydantic_model, BaseModel):
        return str(pydantic_model)

    model_dict = {}
    for name, field in pydantic_model.__annotations__.items():
        origin = get_origin(field)
        if origin is Union:
            args = get_args(field)
            if type(None) in args:
                # Remove 'None' type from Union type
                args = tuple(arg for arg in args if arg is not type(None))  # noqa: E721
            if len(args) == 1:
                field = args[0]
                origin = get_origin(field)
        if origin is dict or origin is list:
            model_dict[name] = str(field)
        else:
            model_dict[name] = create_nested_dict(field)
    return model_dict


if __name__ == "__main__":
    from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
        MediapipeModelInfo,
    )
    from freemocap.data_layer.data_saver import SkeletonSchema

    skeleton_schema = SkeletonSchema(schema_dict=MediapipeModelInfo.skeleton_schema)
    print(create_nested_dict(skeleton_schema))



import logging
import multiprocessing
import shutil
import threading
from pathlib import Path
from typing import Callable

from PySide6.QtCore import Signal, Slot, Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QPushButton, QVBoxLayout, QWidget, QGroupBox
from pyqtgraph.parametertree import Parameter, ParameterTree

from freemocap.data_layer.recording_models.post_processing_parameter_models import (
    ProcessingParameterModel,
)
from freemocap.data_layer.recording_models.recording_info_model import RecordingInfoModel
from freemocap.gui.qt.utilities.save_and_load_gui_state import GuiState
from freemocap.gui.qt.widgets.control_panel.calibration_control_panel import CalibrationControlPanel
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.parameter_groups.create_parameter_groups import (
    create_mediapipe_parameter_group,
    create_3d_triangulation_parameter_group,
    create_post_processing_parameter_group,
    extract_parameter_model_from_parameter_tree,
    RUN_IMAGE_TRACKING_NAME,
    RUN_3D_TRIANGULATION_NAME,
    RUN_BUTTERWORTH_FILTER_NAME,
    NUMBER_OF_PROCESSES_PARAMETER_NAME,
)
from freemocap.gui.qt.workers.process_motion_capture_data_thread_worker import (
    ProcessMotionCaptureDataThreadWorker,
)

logger = logging.getLogger(__name__)


class ProcessMotionCaptureDataPanel(QWidget):
    processing_finished_signal = Signal()

    def __init__(
        self,
        recording_processing_parameters: ProcessingParameterModel,
        get_active_recording_info: Callable[..., RecordingInfoModel],
        kill_thread_event: threading.Event,
        log_update: Callable,
        gui_state: GuiState,
        parent=None,
    ):
        super().__init__(parent=parent)

        self._kill_thread_event = kill_thread_event
        self._get_active_recording_info = get_active_recording_info
        self._log_update = log_update
        self._recording_processing_parameter_model = recording_processing_parameters
        self._recording_processing_parameter_model.recording_info_model = self._get_active_recording_info()

        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(self._layout)

        calibration_group_box = QGroupBox("Selected Capture Volume Calibration TOML:")
        vbox = QVBoxLayout()
        calibration_group_box.setLayout(vbox)
        self._layout.addWidget(calibration_group_box)

        self._calibration_control_panel = CalibrationControlPanel(
            get_active_recording_info=self._get_active_recording_info,
            kill_thread_event=self._kill_thread_event,
            gui_state=gui_state,
            parent=self,
        )
        vbox.addWidget(self._calibration_control_panel)

        self._process_motion_capture_data_button = QPushButton(
            "Process Motion Capture Videos",
        )
        self._process_motion_capture_data_button.setFont(QFont("Dosis", 16, QFont.Weight.Bold))
        self._process_motion_capture_data_button.setObjectName("process_motion_capture_button")

        self._process_motion_capture_data_button.clicked.connect(self._launch_process_motion_capture_data_thread_worker)
        self._layout.addWidget(self._process_motion_capture_data_button)

        self._parameter_tree_widget = ParameterTree(parent=self, showHeader=False)
        self._layout.addWidget(self._parameter_tree_widget)

        self._parameter_group = self._add_parameters_to_parameter_tree_widget(
            self._parameter_tree_widget,
            session_processing_parameter_model=self._recording_processing_parameter_model,
        )

        self._process_motion_capture_data_thread_worker = None

    @property
    def process_motion_capture_data_button(self):
        return self._process_motion_capture_data_button

    def calibrate_from_active_recording(
        self, charuco_square_size_mm: float, use_charuco_as_groundplane: bool, charuco_board_name: str
    ):
        self._calibration_control_panel.calibrate_from_active_recording(
            charuco_square_size_mm=charuco_square_size_mm,
            use_charuco_as_groundplane=use_charuco_as_groundplane,
            charuco_board_name=charuco_board_name,
        )

    def update_calibration_path(self) -> None:
        self._calibration_control_panel.update_calibrate_from_active_recording_button_text()
        self._calibration_control_panel.update_calibration_toml_path()

    def _add_parameters_to_parameter_tree_widget(
        self,
        parameter_tree_widget: ParameterTree,
        session_processing_parameter_model: ProcessingParameterModel,
    ):
        parameter_group = self._convert_session_processing_parameter_model_to_parameter_group(
            session_processing_parameter_model
        )
        parameter_tree_widget.setParameters(parameter_group, showTop=False)
        parameter_tree_widget.setObjectName("parameter-tree-widget")
        return parameter_group

    def _convert_session_processing_parameter_model_to_parameter_group(
        self,
        session_processing_parameter_model: ProcessingParameterModel,
    ):
        return Parameter.create(
            name="Processing Parameters",
            type="group",
            children=[
                dict(
                    name="2d Image Trackers",
                    type="group",
                    children=[
                        self._create_new_run_this_step_parameter(run_step_name=RUN_IMAGE_TRACKING_NAME),
                        self._create_num_processes_parameter(),
                        create_mediapipe_parameter_group(session_processing_parameter_model.tracking_parameters_model),
                    ],
                    tip="Methods for tracking 2d points in images (e.g. mediapipe, deeplabcut(TODO), openpose(TODO), etc ...)",
                ),
                dict(
                    name="3d Triangulation Methods",
                    type="group",
                    children=[
                        self._create_new_run_this_step_parameter(run_step_name=RUN_3D_TRIANGULATION_NAME),
                        create_3d_triangulation_parameter_group(
                            session_processing_parameter_model.anipose_triangulate_3d_parameters_model
                        ),
                    ],
                    tip="Methods for triangulating 3d points from 2d points (using epipolar geometry and the 'camera_calibration' data).",
                ),
                dict(
                    name="Post Processing (data cleaning)",
                    type="group",
                    children=[
                        self._create_new_run_this_step_parameter(run_step_name=RUN_BUTTERWORTH_FILTER_NAME),
                        create_post_processing_parameter_group(
                            session_processing_parameter_model.post_processing_parameters_model
                        ),
                    ],
                    tip="Methods for cleaning up the data (e.g. filtering/smoothing, gap filling, etc ...)"
                    "TODO - Add/expose more post processing methods here (e.g. gap filling, outlier removal, etc ...)",
                ),
            ],
        )

    def _create_session_parameter_model(
        self,
    ) -> ProcessingParameterModel:
        recording_processing_parameter_model = extract_parameter_model_from_parameter_tree(
            parameter_object=self._parameter_group
        )
        recording_processing_parameter_model.recording_info_model = self._get_active_recording_info()

        if self._calibration_control_panel.calibration_toml_path:
            recording_processing_parameter_model.recording_info_model.calibration_toml_path = (
                self._calibration_control_panel.calibration_toml_path
            )
        else:
            recording_processing_parameter_model.recording_info_model.calibration_toml_path = (
                self._calibration_control_panel.open_load_camera_calibration_toml_dialog()
            )

        if not recording_processing_parameter_model.recording_info_model.calibration_toml_path:
            logger.error(
                "No calibration TOML selected - Processing will fail at the '3d triangulation' step (but it will get through the 'Tracking things in 2d images' step)."
            )

        return recording_processing_parameter_model

    def _create_new_run_this_step_parameter(self, run_step_name: str):
        parameter = Parameter.create(
            name=run_step_name,
            type="bool",
            value=True,
            tip="If you have already run this step, you can skip it." "re-running it will overwrite the existing data.",
        )
        parameter.sigValueChanged.connect(self.enable_this_parameter_group)

        return parameter

    def _create_num_processes_parameter(self):
        parameter = Parameter.create(
            name=NUMBER_OF_PROCESSES_PARAMETER_NAME,
            type="int",
            limits=(1, multiprocessing.cpu_count() - 1),
            value=(multiprocessing.cpu_count() - 1),
            tip="If your computer has issues processing larger videos, you can try setting number of processes to 1 to process each video serially.",
        )
        return parameter

    def enable_this_parameter_group(self, parameter):
        run_this_step_bool = parameter.value()
        parameter_group = parameter.parent()
        for child in parameter_group.children():
            if child.name().split(" ")[0] != "Run":
                logger.debug(
                    f"{'Enabling' if run_this_step_bool else 'Disabling'} {child.name()} in processing pipeline"
                )
                self.set_parameter_enabled(child, run_this_step_bool)

    def set_parameter_enabled(self, parameter, enabled_bool):
        parameter.setOpts(enabled=enabled_bool)
        if parameter.hasChildren():
            for child in parameter.children():
                self.set_parameter_enabled(child, enabled_bool)

    def _launch_process_motion_capture_data_thread_worker(self):
        logger.debug("Launching process motion capture data process")
        session_parameter_model = self._create_session_parameter_model()
        session_parameter_model.recording_info_model = self._get_active_recording_info()

        if session_parameter_model.recording_info_model is None:
            logger.error("No active recording selected.")
            return
        if not Path(session_parameter_model.recording_info_model.path).exists():
            logger.error(f"Recording path does not exist: {session_parameter_model.recording_info_model.path}.")
            return

        if session_parameter_model.recording_info_model.single_video_check is False:
            selected_camera_calibration_toml_path = self._calibration_control_panel.calibration_toml_path
            session_parameter_model.recording_info_model.calibration_toml_path = selected_camera_calibration_toml_path

            # check if there is already a calibration toml  in the recording folder, if not save this one there
            if (
                len(list(Path(session_parameter_model.recording_info_model.path).glob("*.toml"))) == 0
                and Path(selected_camera_calibration_toml_path).exists()
            ):
                # copy the calibration toml to the recording folder (keeping the original filename)
                logger.info(
                    f"Copying calibration toml from {selected_camera_calibration_toml_path} to {session_parameter_model.recording_info_model.path}"
                )
                Path(session_parameter_model.recording_info_model.path).mkdir(parents=True, exist_ok=True)
                copied_toml_path = (
                    Path(session_parameter_model.recording_info_model.path)
                    / Path(selected_camera_calibration_toml_path).name
                )
                shutil.copyfile(selected_camera_calibration_toml_path, copied_toml_path)

        # set active tracker in recording model to the currently selected tracker
        session_parameter_model.recording_info_model.active_tracker = session_parameter_model.tracking_model_info.name

        self._process_motion_capture_data_thread_worker = ProcessMotionCaptureDataThreadWorker(
            session_parameter_model, kill_event=self._kill_thread_event
        )
        self._process_motion_capture_data_thread_worker.in_progress.connect(self._log_update)
        self._process_motion_capture_data_thread_worker.start()

        self._process_motion_capture_data_thread_worker.finished.connect(self._handle_finished_signal)

    @Slot(bool)
    def _handle_finished_signal(self, successful: bool):
        if successful:
            logger.debug("Process motion capture data process successful.")
            self.processing_finished_signal.emit()
        else:
            logger.debug("Process motion capture data process failed.")


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    window = ProcessMotionCaptureDataPanel()
    window.show()
    app.exec()



import numpy as np
from skellyforge.freemocap_utils.postprocessing_widgets.postprocessing_functions import interpolate_data


class CharucoVisibilityError(RuntimeError):
    """Raised when no frame satisfies the ‘all-corners-visible & stationary’ criteria."""


class CharucoVelocityError(RuntimeError):
    """Raised when the velocity of the ChArUco corners is too high to be considered stationary."""


def get_charuco_x_and_y_idx(number_of_squares_width: int, number_of_squares_height: int):
    """
    For a given board definition, get the corner marker indexes needed to make the x and y vectors
    """

    num_cols = number_of_squares_width - 1  # corner columns
    num_rows = number_of_squares_height - 1  # corner rows

    idx_x = num_cols * (num_rows - 1)
    idx_y = num_cols - 1

    return idx_x, idx_y


def get_unit_vector(vector: np.ndarray) -> np.ndarray:
    return vector / np.linalg.norm(vector)


def compute_basis_vectors_of_new_reference(
        charuco_frame: np.ndarray, number_of_squares_width: int, number_of_squares_height: int
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    origin = charuco_frame[0]

    idx_x, idx_y = get_charuco_x_and_y_idx(
        number_of_squares_width=number_of_squares_width, number_of_squares_height=number_of_squares_height
    )

    x_vec = charuco_frame[idx_x] - origin
    y_vec = charuco_frame[idx_y] - origin

    x_hat = get_unit_vector(x_vec)
    y_hat_raw = get_unit_vector(y_vec)
    z_hat = get_unit_vector(np.cross(x_hat, y_hat_raw))
    y_hat = get_unit_vector(np.cross(z_hat, x_hat))

    return x_hat, y_hat, z_hat


def find_still_frame(points_velocity: np.ndarray, max_allowed_velocity: float = 1.0) -> int:
    visible_frames = ~np.isnan(points_velocity).any(axis=1)
    if not np.any(visible_frames):
        raise CharucoVisibilityError(
            "No frame found where all three required ChArUco corners are on the ground and visible."
        )

    max_velocity_per_frame = np.nanmax(points_velocity[visible_frames], axis=1)
    if np.nanmin(max_velocity_per_frame) > max_allowed_velocity:
        raise CharucoVelocityError(
            f"All frames have ChArUco corner velocity > {max_allowed_velocity:.2f} mm/s — check that the board is stationary."
        )

    best_visible_index = int(np.nanargmin(max_velocity_per_frame))
    best_velocity_index = np.where(visible_frames)[0][
        best_visible_index
    ]  # maps back to the original array (which we sliced earlier to get rid of rows with nans)
    return best_velocity_index


def find_good_frame(
        charuco_data: np.ndarray,
        number_of_squares_width: int,
        number_of_squares_height: int,
        frame_to_use: int = 0,
        search_range: int = 120,
):
    if frame_to_use == 0:
        slice_to_search = slice(0, search_range)
    elif frame_to_use == -1:
        slice_to_search = slice(-search_range, None)
    elif frame_to_use > 0:
        start_frame = max(0, frame_to_use - search_range)
        end_frame = min(charuco_data.shape[0], frame_to_use + search_range)
        slice_to_search = slice(start_frame, end_frame)
    else:
        raise ValueError(f"Invalid value for frame_to_use: {frame_to_use}")

    idx_x, idx_y = get_charuco_x_and_y_idx(
        number_of_squares_width=number_of_squares_width, number_of_squares_height=number_of_squares_height
    )

    charuco_corners = charuco_data[slice_to_search, [0, idx_y, idx_x]]
    charuco_corners_velocity = np.linalg.norm(np.diff(charuco_corners, axis=0), axis=2)
    best_velocity_frame = find_still_frame(points_velocity=charuco_corners_velocity)

    frame_offset = slice_to_search.start or 0
    best_position_frame = best_velocity_frame + frame_offset + 1
    return best_position_frame


def skellyforge_data(raw_charuco_data: np.ndarray):
    interpolated_data = interpolate_data.interpolate_skeleton_data(
        skeleton_data=raw_charuco_data
    )
    return interpolated_data



import logging
from pathlib import Path
from typing import Tuple, Union

import numpy as np
from matplotlib import pyplot as plt

from freemocap.core_processes.capture_volume_calibration.save_3d_data_to_npy import (
    save_3d_data_to_npy,
)
from freemocap.core_processes.capture_volume_calibration.triangulate_3d_data import triangulate_3d_data
from freemocap.data_layer.recording_models.post_processing_parameter_models import ProcessingParameterModel

logger = logging.getLogger(__name__)


def run_reprojection_error_filtering(
    image_data_numCams_numFrames_numTrackedPts_XYZ: np.ndarray,
    raw_skel3d_frame_marker_xyz: np.ndarray,
    skeleton_reprojection_error_cam_fr_mar: np.ndarray,
    skeleton_reprojection_error_fr_mar: np.ndarray,
    anipose_calibration_object,
    processing_parameters: ProcessingParameterModel,
) -> np.ndarray:
    """
    Runs reprojection error filtering on 3d data, saves the filtered 3d data and reprojection error data, and creates and saves a debug plot.

    :param image_data_numCams_numFrames_numTrackedPts_XYZ: original 2d data
    :param raw_skel3d_frame_marker_xyz: original 3d data
    :param skeleton_reprojection_error_cam_fr_mar: original per camerareprojection error
    :param skeleton_reprojection_error_fr_mar: original per frame average reprojection error
    :param anipose_calibration_object: anipose calibration object
    :param processing_parameters: processing parameters

    :return: filtered 3d data
    """
    if hasattr(
        processing_parameters.tracking_model_info, "num_tracked_points_body"
    ):  # we don't want to reproject hand and face data
        num_tracked_points = processing_parameters.tracking_model_info.num_tracked_points_body
    else:
        num_tracked_points = processing_parameters.tracking_model_info.num_tracked_points

    (
        reprojection_filtered_skel3d_frame_marker_xyz,
        reprojection_filtered_skeleton_reprojection_error_fr_mar,
        reprojection_filtered_skeleton_reprojection_error_cam_fr_mar,
    ) = filter_by_reprojection_error(
        reprojection_error_camera_frame_marker=skeleton_reprojection_error_cam_fr_mar,
        reprojection_error_frame_marker=skeleton_reprojection_error_fr_mar,
        reprojection_error_confidence_threshold=processing_parameters.anipose_triangulate_3d_parameters_model.reprojection_error_confidence_cutoff,
        image_2d_data=image_data_numCams_numFrames_numTrackedPts_XYZ[:, :, :, :2],
        raw_skel3d_frame_marker_xyz=raw_skel3d_frame_marker_xyz,
        anipose_calibration_object=anipose_calibration_object,
        num_tracked_points=num_tracked_points,
        use_triangulate_ransac=processing_parameters.anipose_triangulate_3d_parameters_model.use_triangulate_ransac_method,
        minimum_cameras_to_reproject=processing_parameters.anipose_triangulate_3d_parameters_model.minimum_cameras_to_reproject,
    )
    save_3d_data_to_npy(
        data3d_numFrames_numTrackedPoints_XYZ=reprojection_filtered_skel3d_frame_marker_xyz,
        data3d_numFrames_numTrackedPoints_reprojectionError=reprojection_filtered_skeleton_reprojection_error_fr_mar,
        data3d_numCams_numFrames_numTrackedPoints_reprojectionError=reprojection_filtered_skeleton_reprojection_error_cam_fr_mar,
        path_to_folder_where_data_will_be_saved=processing_parameters.recording_info_model.raw_data_folder_path,
        processing_level="reprojection_filtered",
        file_prefix=processing_parameters.tracking_model_info.name,
    )
    plot_reprojection_error(
        raw_reprojection_error_frame_marker=skeleton_reprojection_error_fr_mar,
        filtered_reprojection_error_frame_marker=reprojection_filtered_skeleton_reprojection_error_fr_mar,
        reprojection_error_threshold=float(
            np.nanpercentile(
                skeleton_reprojection_error_cam_fr_mar,
                processing_parameters.anipose_triangulate_3d_parameters_model.reprojection_error_confidence_cutoff,
            )
        ),
        output_folder_path=processing_parameters.recording_info_model.raw_data_folder_path,
    )
    return reprojection_filtered_skel3d_frame_marker_xyz


def filter_by_reprojection_error(
    reprojection_error_camera_frame_marker: np.ndarray,
    reprojection_error_frame_marker: np.ndarray,
    reprojection_error_confidence_threshold: float,
    image_2d_data: np.ndarray,
    raw_skel3d_frame_marker_xyz: np.ndarray,
    anipose_calibration_object,
    num_tracked_points: int,
    use_triangulate_ransac: bool = False,
    minimum_cameras_to_reproject: int = 3,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    total_cameras = image_2d_data.shape[0]
    num_cameras_to_remove = 1

    if total_cameras <= minimum_cameras_to_reproject:
        logger.warning(
            f"Not enough cameras to filter by reprojection error. There are {total_cameras} cameras, but minimum number of cameras is {minimum_cameras_to_reproject}. Returning unfiltered data."
        )
        return (
            raw_skel3d_frame_marker_xyz,
            reprojection_error_frame_marker,
            reprojection_error_camera_frame_marker,
        )

    body2d_camera_frame_marker_xy = image_2d_data[:, :, :num_tracked_points, :2]
    bodyReprojErr_camera_frame_marker = reprojection_error_camera_frame_marker[:, :, :num_tracked_points]

    reprojection_error_confidence_threshold = min(max(reprojection_error_confidence_threshold, 0), 100)

    reprojection_error_threshold = np.nanpercentile(
        np.nanmean(bodyReprojErr_camera_frame_marker, axis=0), reprojection_error_confidence_threshold, method="weibull"
    )  # TODO: try running this on reprojection_error_frame_marker with body points pulled out, rather than with cameras included
    logger.info(f"Using reprojection error threshold of {reprojection_error_threshold}")

    data_to_reproject_camera_frame_marker_xy, unique_frame_marker_list = _get_data_to_reproject(
        num_cameras_to_remove=num_cameras_to_remove,
        reprojection_error_threshold=reprojection_error_threshold,
        reprojError_cam_frame_marker=bodyReprojErr_camera_frame_marker,
        input_2d_data_camera_frame_marker_xy=body2d_camera_frame_marker_xy,
    )

    while len(unique_frame_marker_list) > 0 and total_cameras - num_cameras_to_remove >= minimum_cameras_to_reproject:
        logger.info("Retriangulating data")
        (
            retriangulated_data_frame_marker_xyz,
            new_reprojection_error_flat,
            new_reprojError_cam_frame_marker,
        ) = triangulate_3d_data(
            anipose_calibration_object=anipose_calibration_object,
            image_2d_data=data_to_reproject_camera_frame_marker_xy,
            use_triangulate_ransac=use_triangulate_ransac,
        )

        num_cameras_to_remove += 1

        data_to_reproject_camera_frame_marker_xy, unique_frame_marker_list = _get_data_to_reproject(
            num_cameras_to_remove=num_cameras_to_remove,
            reprojection_error_threshold=reprojection_error_threshold,
            reprojError_cam_frame_marker=new_reprojError_cam_frame_marker,
            input_2d_data_camera_frame_marker_xy=data_to_reproject_camera_frame_marker_xy,
        )

    # nan remaining data above threshold
    _nan_data_above_threshold(unique_frame_marker_list, retriangulated_data_frame_marker_xyz)

    # put retriangulated data back in place
    filtered_skel3d_frame_marker_xyz = raw_skel3d_frame_marker_xyz.copy()
    filtered_skel3d_frame_marker_xyz[:, :num_tracked_points, :] = retriangulated_data_frame_marker_xyz

    filtered_reprojection_error_frame_marker = reprojection_error_frame_marker.copy()
    filtered_reprojection_error_frame_marker[:, :num_tracked_points] = new_reprojection_error_flat

    filtered_reprojection_error_camera_frame_marker = reprojection_error_camera_frame_marker.copy()
    filtered_reprojection_error_camera_frame_marker[:, :, :num_tracked_points] = new_reprojError_cam_frame_marker

    return (
        filtered_skel3d_frame_marker_xyz,
        filtered_reprojection_error_frame_marker,
        filtered_reprojection_error_camera_frame_marker,
    )


def _nan_data_above_threshold(unique_frame_marker_list: list, retriangulated_data_frame_marker_xyz: np.ndarray) -> None:
    if len(unique_frame_marker_list) > 0:
        logger.info(
            f"Out of cameras to remove, setting {len(unique_frame_marker_list)} points with reprojection error above threshold to NaNs"
        )
        for frame_marker in unique_frame_marker_list:
            retriangulated_data_frame_marker_xyz[frame_marker[0], frame_marker[1], :] = np.nan


def _get_data_to_reproject(
    num_cameras_to_remove: int,
    reprojection_error_threshold: float,
    reprojError_cam_frame_marker: np.ndarray,
    input_2d_data_camera_frame_marker_xy: np.ndarray,
) -> tuple[np.ndarray, list]:
    indices_above_threshold = np.nonzero(reprojError_cam_frame_marker > reprojection_error_threshold)
    logger.debug(f"SHAPE OF INDICES ABOVE THRESHOLD: {indices_above_threshold[0].shape}")

    total_frame_marker_combos = (
        input_2d_data_camera_frame_marker_xy.shape[1] * input_2d_data_camera_frame_marker_xy.shape[2]
    )
    unique_frame_marker_list = _get_unique_frame_marker_list(indices_above_threshold=indices_above_threshold)
    logger.info(
        f"number of frame/marker combos with reprojection error above threshold: {len(unique_frame_marker_list)} ({len(unique_frame_marker_list) / total_frame_marker_combos * 100:.1f} percent of total)"
    )

    cameras_to_remove, frames_to_reproject, markers_to_reproject = _get_camera_frame_marker_lists_to_reproject(
        reprojError_cam_frame_marker=reprojError_cam_frame_marker,
        frame_marker_list=unique_frame_marker_list,
        num_cameras_to_remove=num_cameras_to_remove,
    )

    data_to_reproject_camera_frame_marker_xy = _set_unincluded_data_to_nans(
        input_2d_data=input_2d_data_camera_frame_marker_xy,
        frames_with_reprojection_error=frames_to_reproject,
        markers_with_reprojection_error=markers_to_reproject,
        cameras_to_remove=cameras_to_remove,
    )

    return (data_to_reproject_camera_frame_marker_xy, unique_frame_marker_list)


def _get_unique_frame_marker_list(
    indices_above_threshold: np.ndarray,
) -> list:
    return list(set(zip(indices_above_threshold[1], indices_above_threshold[2])))


def _get_camera_frame_marker_lists_to_reproject(
    reprojError_cam_frame_marker: np.ndarray,
    frame_marker_list: list,
    num_cameras_to_remove: int,
) -> Tuple[list, list, list]:
    """
    Generate the lists of cameras, frames, and markers to reproject based on the given input.
    Find the cameras with the worst reprojection errors for the given frames and markers.
    Args:
        reprojError_cam_frame_marker (np.ndarray): The array containing the reprojection errors for each camera, frame, and marker.
        frame_marker_list (list): The list of tuples containing the frame and marker indices.
        num_cameras_to_remove (int): The number of cameras to remove.
    Returns:
        Tuple[list, list, list]: A tuple containing the lists of cameras to remove, frames to reproject, and markers to reproject.
    """
    cameras_to_remove = []
    frames_to_reproject = []
    markers_to_reproject = []
    for frame, marker in frame_marker_list:
        frames_to_reproject.append(frame)
        markers_to_reproject.append(marker)
        max_indices = reprojError_cam_frame_marker[:, frame, marker].argsort()[::-1][:num_cameras_to_remove]
        cameras_to_remove.append(list(max_indices))
    return (cameras_to_remove, frames_to_reproject, markers_to_reproject)


def _set_unincluded_data_to_nans(
    input_2d_data: np.ndarray,
    frames_with_reprojection_error: list,
    markers_with_reprojection_error: list,
    cameras_to_remove: list[list[int]],
) -> np.ndarray:
    data_to_reproject = input_2d_data.copy()
    for list_of_cameras, frame, marker in zip(
        cameras_to_remove, frames_with_reprojection_error, markers_with_reprojection_error
    ):
        for camera in list_of_cameras:
            data_to_reproject[camera, frame, marker, :] = np.nan
    return data_to_reproject


def plot_reprojection_error(
    raw_reprojection_error_frame_marker: np.ndarray,
    filtered_reprojection_error_frame_marker: np.ndarray,
    reprojection_error_threshold: float,
    output_folder_path: Union[str, Path],
) -> None:
    title = "Mean Reprojection Error Per Frame"
    file_name = "debug_reprojection_error_filtering.png"
    output_filepath = Path(output_folder_path) / file_name
    raw_mean_reprojection_error_per_frame = np.nanmean(
        raw_reprojection_error_frame_marker,
        axis=1,
    )
    filtered_mean_reprojection_error_per_frame = np.nanmean(
        filtered_reprojection_error_frame_marker,
        axis=1,
    )
    plt.plot(raw_mean_reprojection_error_per_frame, color="blue", label="Data Before Filtering")
    plt.plot(filtered_mean_reprojection_error_per_frame, color="orange", alpha=0.9, label="Data After Filtering")
    plt.xlabel("Frame")
    plt.ylabel("Mean Reprojection Error Across Markers (mm)")
    plt.ylim(0, 2 * reprojection_error_threshold)
    plt.hlines(
        y=reprojection_error_threshold,
        xmin=0,
        xmax=len(raw_mean_reprojection_error_per_frame),
        color="red",
        label="Cutoff Threshold",
    )
    plt.title(title)
    plt.legend(loc="upper right")
    logger.info(f"Saving debug plots to: {output_filepath}")
    plt.savefig(output_filepath, dpi=300)



import json
import logging
from pathlib import Path
from typing import Tuple, Union

import numpy as np
import pandas as pd
from skellytracker.trackers.base_tracker.model_info import ModelInfo
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
    MediapipeModelInfo,
)

from freemocap.data_layer.data_saver.data_loader import DataLoader
from freemocap.data_layer.data_saver.data_models import InfoDict

logger = logging.getLogger(__name__)

from typing import Any, Dict

mediapipe_model_info = MediapipeModelInfo()


class DataSaver:
    def __init__(
        self,
        recording_folder_path: Union[Path, str],
        include_hands: bool = True,
        include_face: bool = True,
        model_info: ModelInfo = mediapipe_model_info,
    ):
        """
        Initialize DataFrameManager with the given recording_folder_path.

        Args:
            recording_folder_path (Union[Path, str]): path to the folder containing the recording data.
            include_hands (bool): flag to include hands data in the processing.
            include_face (bool): flag to include face data in the processing.
        """
        self.recording_folder_path = Path(recording_folder_path)
        self._recording_name = self.recording_folder_path.name

        self._data_loader = DataLoader(
            recording_folder_path=self.recording_folder_path,
            include_hands=include_hands,
            include_face=include_face,
            model_info=model_info,
        )

        self.recording_data_by_frame = None
        self.number_of_frames = None
        self.model_info = model_info

    def save_all(self):
        """
        Load all data, validate it, and create the recording_data_by_frame_number dictionary.

        Returns:
            Dict[int, Dict]: a dictionary containing the processed data, indexed by frame number.
        """

        self.recording_data_by_frame = self._data_loader.get_data_by_frame()

        self.save_to_json()
        self.save_to_csv()
        self.save_to_npy()
        self.save_to_tidy_csv()

    def save_to_json(self, save_path: Union[str, Path] = None):
        dict_to_save = {}
        dict_to_save["info"] = self._get_info_dict().model_dump()
        dict_to_save["data_by_frame"] = self.recording_data_by_frame

        if save_path is None:
            save_path = Path(self.recording_folder_path) / f"{self._recording_name}_by_frame.json"

        logger.info(f"Saving recording data to {save_path}")

        save_path.write_text(json.dumps(dict_to_save, indent=4), encoding="utf-8")

    def save_to_csv(self, save_path: Union[str, Path] = None):
        data_for_dataframe = []

        for frame_data in self.recording_data_by_frame.values():
            data_for_dataframe.append(self._generate_frame_data_row(frame_data))

        # Create DataFrame and save to csv
        df = pd.DataFrame(data_for_dataframe)

        if save_path is None:
            save_path = self.recording_folder_path / f"{self._recording_name}_by_trajectory.csv"

        df.to_csv(save_path, index=False)
        logger.info(f"Saved recording data to {save_path}")

    def _parse_keypoint_name(self, keypoint_name: str) -> Tuple[str, str]:
        if keypoint_name.startswith("right_hand"):
            split_point_name = keypoint_name.split("right_hand_", maxsplit=1)
            model = f"{self.model_info.name}_hand"
            keypoint = f"right_{split_point_name[1]}"
        elif keypoint_name.startswith("left_hand"):
            split_point_name = keypoint_name.split("left_hand_", maxsplit=1)
            model = f"{self.model_info.name}_hand"
            keypoint = f"left_{split_point_name[1]}"
        else:
            split_point_name = keypoint_name.split("_", maxsplit=1)
            if len(split_point_name) != 2:
                model = self.model_info.name
                keypoint = keypoint_name
            else:
                model = f"{self.model_info.name}_{split_point_name[0]}"
                keypoint = split_point_name[1]

        return model, keypoint

    def save_to_tidy_csv(self, save_path: Union[str, Path, None] = None):
        tidy_data = []
        # Iterate over frames and tracked points
        for frame_number, frame_data in self.recording_data_by_frame.items():
            timestamp = frame_data["timestamps"]["mean"]
            timestamp_by_camera = frame_data["timestamps"]["by_camera"]

            for point_name, coordinates in frame_data["tracked_points"].items():
                model, keypoint = self._parse_keypoint_name(point_name)
                tidy_data.append(
                    {
                        "frame": frame_number,
                        "timestamp": timestamp,
                        "timestamp_by_camera": timestamp_by_camera,
                        "model": model,
                        "keypoint": keypoint,
                        "x": coordinates.get("x", None),
                        "y": coordinates.get("y", None),
                        "z": coordinates.get("z", None),
                        "reprojection_error": frame_data["reprojection_error"].get(keypoint, {}).get("value", None),
                    }
                )

        # Convert to DataFrame
        tidy_df = pd.DataFrame(tidy_data)

        # Set the save path if not provided
        if save_path is None:
            save_path = self.recording_folder_path / f"{self._recording_name}_by_frame.csv"

        # Save the DataFrame to a CSV file
        tidy_df.to_csv(save_path, index=False)
        logger.info(f"Saved tidy recording data to {save_path}")

    def save_to_npy(self, save_path: Union[str, Path] = None):
        if save_path is None:
            save_path = self.recording_folder_path / f"{self._recording_name}_frame_name_xyz.npy"
        np.save(save_path, self._data_loader.data_frame_name_xyz)
        logger.info(f"Saved recording data to {save_path}")

    def _generate_frame_data_row(self, frame_data: Dict[str, Any]) -> Dict:
        """
        Generate a data row for a given frame number.
        """
        frame_data_row = {}

        frame_data_row["timestamp"] = frame_data["timestamps"]["mean"]
        frame_data_row["timestamp_by_camera"] = str(frame_data["timestamps"]["by_camera"])

        for key, tracked_point in frame_data["tracked_points"].items():
            frame_data_row[f"{key}_x"] = tracked_point["x"]
            frame_data_row[f"{key}_y"] = tracked_point["y"]
            frame_data_row[f"{key}_z"] = tracked_point["z"]

        return frame_data_row

    def _get_info_dict(self):
        return InfoDict(
            # segment_lengths=self._data_loader.segment_lengths,
            schemas=self._data_loader._model_info.segment_connections,
        )


if __name__ == "__main__":
    # recording_data_saver = DataSaver(recording_folder_path=get_sample_data_path())
    recording_data_saver = DataSaver(
        recording_folder_path=r"C:\Users\aaron\FreeMocap_Data\recording_sessions\freemocap_sample_data"
    )
    recording_data_by_frame_number = recording_data_saver.save_all()



from src.api.routes.camera.cam_data_ws import cam_ws_router
from src.api.routes.camera.camera_route import camera_router
from src.api.routes.health.health_check_route import healthcheck_router
from src.api.routes.home.home import home_router
from src.api.routes.session.session_router import session_router
from src.api.routes.startup.startup import startup_router

# REGISTER NEW ROUTES HERE
enabled_routers = [
    healthcheck_router,
    camera_router,
    cam_ws_router,
    startup_router,
    home_router,
    session_router,
]



from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QDockWidget,
    QLabel,
    QSplitter,
    QVBoxLayout,
    QWidget,
)
from skellycam import SkellyCamParameterTreeWidget

from freemocap.gui.qt.widgets.control_panel.calibration_control_panel import (
    CalibrationControlPanel,
)
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.process_motion_capture_data_panel import (
    ProcessMotionCaptureDataPanel,
)


class ControlPanelDockWidget(QDockWidget):
    def __init__(
            self,
            camera_configuration_parameter_tree_widget: SkellyCamParameterTreeWidget,
            calibration_control_panel: CalibrationControlPanel,
            process_motion_capture_data_panel: ProcessMotionCaptureDataPanel,
            visualize_data_widget: QWidget,
            parent=None,
    ):
        super().__init__(parent=parent)

        self._splitter = QSplitter()
        self._splitter.setOrientation(Qt.Orientation.Vertical)
        self.setWidget(self._splitter)

        self._add_widget_to_splitter(
            widget=camera_configuration_parameter_tree_widget,
            title_str="Camera Configuration",
        )

        self._add_widget_to_splitter(
            widget=calibration_control_panel,
            title_str="Capture Volume Calibration",
        )

        self._add_widget_to_splitter(
            widget=process_motion_capture_data_panel,
            title_str="Process Motion Capture Data",
        )

        self._add_widget_to_splitter(
            widget=visualize_data_widget,
            title_str="Visualize Motion Capture Data",
            add_stretch=True,
        )

    def _add_widget_to_splitter(self, widget: QWidget, title_str: str, add_stretch: bool = False):
        dummy_widget = QWidget()
        layout = QVBoxLayout()
        dummy_widget.setLayout(layout)
        layout.addWidget(self._create_section_title(title=title_str))
        layout.addWidget(widget)
        if add_stretch:
            layout.addStretch()

        self._splitter.addWidget(dummy_widget)

    def _create_section_title(self, title: str):
        label = QLabel(title)
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        label.setAlignment(Qt.AlignmentFlag.AlignTop)
        label.setStyleSheet("font-size: 20px; font-weight: bold;")
        return label



from pathlib import Path
from typing import Union

from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.anipose_camera_calibrator import (
    AniposeCameraCalibrator,
)
from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)


def headless_calibration(
        path_to_folder_of_calibration_videos: Path,
        charuco_board_object=CharucoBoardDefinition,
        charuco_square_size: Union[int, float] = 39,
        pin_camera_0_to_origin: bool = True,
):
    anipose_camera_calibrator = AniposeCameraCalibrator(
        charuco_board_object=charuco_board_object,
        charuco_square_size=charuco_square_size,
        calibration_videos_folder_path=path_to_folder_of_calibration_videos,
        progress_callback=lambda *args, **kwargs: None,
        # the empty callable is needed, otherwise calibration will cause an error
    )

    anipose_camera_calibrator.calibrate_camera_capture_volume(pin_camera_0_to_origin=pin_camera_0_to_origin)


if __name__ == "__main__":
    path_to_folder_of_calibration_videos = Path(r"C:\Users\aaron\freemocap_data\recording_sessions\freemocap_test_data\synchronized_videos")
    charuco_square_size = 58  # size of a black square on your charuco board in mm

    headless_calibration(
        path_to_folder_of_calibration_videos=path_to_folder_of_calibration_videos,
        charuco_square_size=charuco_square_size,
    )
